from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Tuple
from typing import Union, cast
import sys
from dataclasses import dataclass
from inspect import signature
from pathlib import Path
from contextlib import contextmanager
import json
import copy

from .parser import ArgumentParser, HelpFormatter
from .util import Arg, ArgparseArg, get_arg, join_strings, format_type, format_table
from .util import format_arg_help, expand_error_subclasses, SimpleFrozenDict
from .util import CommandNotFoundError, CliParserError, CommandExistsError
from .util import ConverterType, ConvertersType, ErrorHandlersType, StaticCommand
from .util import StaticData, DEFAULT_CONVERTERS, DEFAULT_PLACEHOLDER


_CallableT = TypeVar("_CallableT", bound=Callable)
DEFAULT_EXTRA_KEY = "_extra"


@dataclass
class Command:
    name: str
    func: Callable
    args: List[ArgparseArg]
    description: Optional[str] = None
    allow_extra: bool = False
    parent: Optional[str] = None
    is_placeholder: bool = False

    @property
    def display_name(self) -> str:
        return f"{self.parent} {self.name}" if self.parent else self.name

    def to_static_json(self) -> StaticCommand:
        """Convert the command to a JSON-serializable dict."""
        return {
            "name": self.name,
            "args": [arg.to_static_json() for arg in self.args],
            "description": self.description,
            "allow_extra": self.allow_extra,
            "parent": self.parent,
            "is_placeholder": self.is_placeholder,
        }

    @classmethod
    def from_static_json(
        cls,
        data: StaticCommand,
        converters: ConvertersType = SimpleFrozenDict(),
    ) -> "Command":
        """Initialize the static command from a JSON-serializable dict."""
        return cls(
            name=data["name"],
            func=lambda *args, **kwargs: None,  # dummy function for static use
            args=[
                ArgparseArg.from_static_json(arg, converters) for arg in data["args"]
            ],
            description=data["description"],
            allow_extra=data["allow_extra"],
            parent=data["parent"],
            is_placeholder=data["is_placeholder"],
        )

    @classmethod
    def from_function(
        cls,
        name: str,
        args: Dict[str, Arg],
        func: Callable,
        *,
        parent: Optional[str] = None,
        allow_extra: bool = False,
        extra_key: str = DEFAULT_EXTRA_KEY,
        converters: ConvertersType = SimpleFrozenDict(),
    ) -> "Command":
        """Create a command from a function and its argument annotations."""
        sig = signature(func)
        sig_types = {}
        sig_defaults = {}
        for param_name, param_value in sig.parameters.items():
            annot = param_value.annotation
            if param_name == extra_key:
                annot = List[str]  # set automatically since we know it
            elif annot == param_value.empty:
                annot = str  # default to string for unset types
            sig_types[param_name] = annot
            sig_defaults[param_name] = (
                param_value.default
                if param_value.default != param_value.empty
                else DEFAULT_PLACEHOLDER  # placeholder for unset defaults
            )
            if param_name not in args:  # support args not in decorator
                args[param_name] = Arg()
        cli_args = []
        for param, arg_info in args.items():
            if param not in sig_types:  # unknown argument
                path = join_strings(parent, name)
                err = f"argument not found in function for '{path}': {param}"
                raise CliParserError(err)

            def get_converter(arg_type: Type) -> Optional[ConverterType]:
                return converters.get(arg_type, arg_info.converter)

            param_type = sig_types[param]
            converter = get_converter(param_type)
            arg_type = converter or param_type
            arg = get_arg(
                param,
                arg_info,
                arg_type,
                orig_type=param_type,
                default=sig_defaults[param],
                skip_resolve=converter is not None,
                get_converter=get_converter,
                has_converter=converter is not None,
            )
            arg.help = join_strings(arg.help, f"({format_type(arg.display_type)})")
            cli_args.append(arg)
        return cls(
            name=name,
            func=func,
            args=cli_args,
            description=func.__doc__,
            allow_extra=allow_extra,
            parent=parent,
        )


class Radicli:
    prog: Optional[str]
    help: str
    version: Optional[str]
    converters: ConvertersType
    extra_key: str
    fill_defaults: bool
    commands: Dict[str, Command]
    subcommands: Dict[str, Dict[str, Command]]
    errors: ErrorHandlersType
    _subcommand_key: str
    _help_arg: str
    _version_arg: str

    def __init__(
        self,
        *,
        prog: Optional[str] = None,
        help: str = "",
        version: Optional[str] = None,
        converters: ConvertersType = SimpleFrozenDict(),
        errors: Optional[ErrorHandlersType] = None,
        extra_key: str = DEFAULT_EXTRA_KEY,
        fill_defaults: bool = True,
    ) -> None:
        """Initialize the CLI and create the registry."""
        self.prog = prog
        self.help = help.strip()
        self.version = version
        self.converters = dict(DEFAULT_CONVERTERS)  # make sure to copy
        self.converters.update(converters)
        self.extra_key = extra_key
        self.fill_defaults = fill_defaults
        self.commands = {}
        self.subcommands = {}
        self.errors = dict(errors) if errors is not None else {}
        self._subcommand_key = "__subcommand__"  # should not conflict with arg name!
        self._help_arg = "--help"
        self._version_arg = "--version"

    # Using underscored argument names here to prevent conflicts if CLI commands
    # define arguments called "name" that are passed in via **args
    def command(self, _name: str, **args: Arg) -> Callable[[_CallableT], _CallableT]:
        """The decorator used to wrap command functions."""
        return self._command(_name, args, self.commands, allow_extra=False)

    def command_with_extra(
        self, _name: str, **args: Arg
    ) -> Callable[[_CallableT], _CallableT]:
        """
        The decorator used to wrap command functions. Supports additional
        arguments passed in as the keyword arg self.extra_key as a list.
        """
        return self._command(_name, args, self.commands, allow_extra=True)

    def subcommand(
        self, _parent: str, _name: str, **args: Arg
    ) -> Callable[[_CallableT], _CallableT]:
        """The decorator used to wrap subcommand functions."""
        return self._subcommand(_parent, _name, args, allow_extra=False)

    def subcommand_with_extra(
        self, _parent: str, _name: str, **args: Arg
    ) -> Callable[[_CallableT], _CallableT]:
        """
        The decorator used to wrap subcommand functions. Supports additional
        arguments passed in as the keyword arg self.extra_key as a list.
        """
        return self._subcommand(_parent, _name, args, allow_extra=True)

    def _subcommand(
        self, parent: str, name: str, args: Dict[str, Arg], *, allow_extra: bool = False
    ) -> Callable[[_CallableT], _CallableT]:
        """The decorator used to wrap subcommands."""
        if parent not in self.subcommands:
            self.subcommands[parent] = {}
        return self._command(
            name, args, self.subcommands[parent], parent=parent, allow_extra=allow_extra
        )

    def _command(
        self,
        name: str,
        args: Dict[str, Arg],
        registry: Dict[str, Command],
        *,
        allow_extra: bool = False,
        parent: Optional[str] = None,
    ) -> Callable[[_CallableT], _CallableT]:
        """The decorator used to wrap command functions."""

        def cli_wrapper(cli_func: _CallableT) -> _CallableT:
            if name in registry:
                raise CommandExistsError(name)
            registry[name] = Command.from_function(
                name,
                args,
                cli_func,
                parent=parent,
                allow_extra=allow_extra,
                extra_key=self.extra_key,
                converters=self.converters,
            )
            return cli_func

        return cli_wrapper

    def placeholder(self, name: str, *, description: Optional[str] = None) -> None:
        """Add empty parent command placeholder with help for subcommands."""
        if name in self.commands:
            raise CommandExistsError(name)

        def func(*args, **kwargs) -> None:
            dummy = Command(
                name=name, func=lambda: None, args=[], description=description
            )
            # If this runs, we want to show the help instead of doing nothing
            self.parse([self._help_arg], dummy, self.subcommands.get(name, {}))

        dummy = Command(
            name=name, func=func, args=[], description=description, is_placeholder=True
        )
        self.commands[name] = dummy

    def call(self, command: Command, args: Optional[List[str]] = None) -> None:
        """Call a single command."""
        run_args = args if args is not None else [*sys.argv[1:]]
        cmd = copy.deepcopy(command)
        cmd.name = ""  # for nicer display in help text
        values = self.parse(run_args, cmd)
        with self.handle_errors():
            command.func(**values)

    def run(self, args: Optional[List[str]] = None) -> None:
        """
        Run the CLI. Should typically be used in the __main__.py nested under a
        `if __name__ == "__main__":` block.
        """
        run_args = args if args is not None else [*sys.argv]
        if len(run_args) <= 1 or run_args[1] == self._help_arg:
            print(self.format_info())
            sys.exit(0)
        # Make single command CLIs available without command name
        if len(self.commands) == 1 and len(self.subcommands) <= 1:
            single_cmd = list(self.commands.keys())[0]
            if run_args[1] != single_cmd:
                run_args.insert(1, single_cmd)
        command = run_args.pop(1)
        args = run_args[1:]
        if self.version and command == self._version_arg:
            print(self.version)
            sys.exit(0)
        subcommands = self.subcommands.get(command, {})
        if command not in self.commands:
            if not subcommands:
                raise CommandNotFoundError(command, list(self.commands))
            # Add a dummy parent to support subcommands without parents
            self.placeholder(command)
        cmd = self.commands[command]
        values = self.parse(args, cmd, subcommands)
        sub = values.pop(self._subcommand_key, None)
        func = subcommands[sub].func if sub else cmd.func
        with self.handle_errors():
            func(**values)

    @contextmanager
    def handle_errors(self):
        # Catch specific error types (and their subclasses), and invoke
        # their handler callback. Handlers can return an integer exit code,
        # which will be passed to sys.exit.
        errors_map = expand_error_subclasses(self.errors)
        try:
            yield
        except tuple(errors_map.keys()) as e:
            handler = errors_map.get(e.__class__)
            if not handler:
                raise e
            err_code = handler(e)
            if err_code is not None:
                sys.exit(err_code)

    def get_parsers(
        self,
        command: Command,
        subcommands: Dict[str, Command] = SimpleFrozenDict(),
    ) -> Tuple[ArgumentParser, Dict[str, Tuple[ArgumentParser, Command]]]:
        """Get parser for a given command."""
        p = ArgumentParser(
            prog=join_strings(self.prog, command.name),
            description=command.description,
            formatter_class=HelpFormatter,
            add_help=not any(a.arg.option == self._help_arg for a in command.args),
            argument_default=DEFAULT_PLACEHOLDER,
        )
        if self.version:
            p.add_argument(self._version_arg, action="version", version=self.version)
        self._add_args(p, command.args)
        subparsers: Dict[str, Tuple[ArgumentParser, Command]] = {}
        if subcommands:
            # We're using the dest to determine whether subcommand was called
            sp = p.add_subparsers(
                title="Subcommands",
                dest=self._subcommand_key,
                parser_class=ArgumentParser,
            )
            for sub_name, sub_cmd in subcommands.items():
                add_help = not any(a.arg.option == self._help_arg for a in sub_cmd.args)
                subp = sp.add_parser(
                    sub_cmd.name,
                    description=sub_cmd.description,
                    help=sub_cmd.description,
                    prog=join_strings(self.prog, sub_cmd.parent, sub_name),
                    add_help=add_help,
                    formatter_class=HelpFormatter,
                    argument_default=DEFAULT_PLACEHOLDER,
                )
                subparsers[sub_cmd.name] = (subp, sub_cmd)
                self._add_args(subp, sub_cmd.args)
        return p, subparsers

    def parse(
        self,
        args: List[str],
        command: Command,
        subcommands: Dict[str, Command] = SimpleFrozenDict(),
        *,
        allow_partial: bool = False,
    ) -> Dict[str, Any]:
        """Parse a list of arguments. Can also be used for testing."""
        p, subparsers = self.get_parsers(command, subcommands)
        # Handling of subcommands is a bit convoluted
        # https://docs.python.org/3/library/argparse.html#sub-commands
        namespace, extra = p.parse_known_args(args)
        values = {**vars(namespace), self.extra_key: extra}
        sub_key = values.pop(self._subcommand_key, None)
        if not sub_key:  # we're not in a subcommand
            return self._validate(command, values, allow_partial=allow_partial)
        if sub_key not in subparsers:
            raise CliParserError(f"invalid subcommand: '{sub_key}'")
        subparser, subcmd = subparsers[cast(str, sub_key)]
        sub_namespace, sub_extra = subparser.parse_known_args(args[1:])
        sub_values = {**vars(sub_namespace), self.extra_key: sub_extra}
        sub_values = self._validate(subcmd, sub_values, allow_partial=allow_partial)
        return {**sub_values, self._subcommand_key: sub_key}

    def _add_args(self, parser: ArgumentParser, args: List[ArgparseArg]) -> None:
        """Add arguments to a parser or subparser."""
        for arg in args:
            if arg.id == self.extra_key:
                continue
            func_args, func_kwargs = arg.to_argparse()
            # Suppress all argument defaults and mark options without defaults
            # as required
            if not self.fill_defaults:
                func_kwargs["default"] = DEFAULT_PLACEHOLDER
                if arg.arg.option:
                    func_kwargs["required"] = arg.default is DEFAULT_PLACEHOLDER
            parser.add_argument(*func_args, **func_kwargs)

    def _validate(
        self, command: Command, values: Dict[str, Any], allow_partial: bool = False
    ) -> Dict[str, Any]:
        """
        Validate required args separately to avoid subparser conflicts.
        Handle extra arguments and raise error if needed. We're doing this
        manually to avoide false positive argparse errors with subcommands.
        """
        extra = values.get(self.extra_key)
        if not command.allow_extra:
            if extra:
                raise CliParserError(f"unrecognized arguments: {' '.join(extra)}")
            values.pop(self.extra_key, None)
        required = []
        for arg in command.args:
            if arg.id not in values or values[arg.id] is DEFAULT_PLACEHOLDER:
                required.append(arg.arg.option or arg.id)
        if required and self.fill_defaults and not allow_partial:
            err = f"the following arguments are required: {', '.join(required)}"
            raise CliParserError(err)
        return values

    def format_info(self) -> str:
        """Nicely format the available command overview and add subcommands."""
        data = []
        for name, cmd in self.commands.items():
            data.append((f"  {name}", format_arg_help(cmd.description)))
            if name in self.subcommands:
                col = f"Subcommands: {', '.join(self.subcommands[name])}"
                data.append(("", col))
        for name in self.subcommands:
            if name not in self.commands:
                col = f"Subcommands: {', '.join(self.subcommands[name])}"
                data.append((f"  {name}", col))
        info = [self.help, "\nAvailable commands:", format_table(data)]
        return join_strings(*info, char="\n")

    def to_static_json(self) -> StaticData:
        """Convert the CLI to a JSON-serializable dict."""
        return {
            "prog": self.prog,
            "help": self.help,
            "version": self.version,
            "extra_key": self.extra_key,
            "commands": {
                cmd.name: cmd.to_static_json() for cmd in self.commands.values()
            },
            "subcommands": {
                parent: {sub.name: sub.to_static_json() for sub in subs.values()}
                for parent, subs in self.subcommands.items()
            },
        }

    def to_static(self, file_path: Union[str, Path]) -> Path:
        """Generate a static representation of the CLI for StaticRadicli."""
        data = self.to_static_json()
        path = Path(file_path)
        with path.open("w", encoding="utf8") as f:
            f.write(json.dumps(data))
        return path



from typing import List, Iterator, Optional, Literal, TypeVar, Generic, Type, Union
from enum import Enum
from dataclasses import dataclass
import pytest
import sys
from contextlib import contextmanager
import tempfile
import shutil
from zipfile import ZipFile
from pathlib import Path
from radicli import Radicli, StaticRadicli, Arg, get_arg, ArgparseArg, Command
from radicli.util import CommandNotFoundError, CliParserError
from radicli.util import ExistingPath, ExistingFilePath, ExistingDirPath
from radicli.util import ExistingFilePathOrDash, DEFAULT_CONVERTERS
from radicli.util import stringify_type, get_list_converter, format_type


@contextmanager
def make_tempdir() -> Iterator[Path]:
    """Run a block in a temp directory and remove it afterwards."""
    d = Path(tempfile.mkdtemp())
    yield d
    shutil.rmtree(str(d))


def test_cli_sys_argv():
    cli = Radicli()
    ran = False

    @cli.command("test")
    def test(a: str, b: int, c: float):
        assert a == "hello"
        assert b == 1
        assert c == 2.0
        nonlocal ran
        ran = True

    sys.argv = ["", "test", "hello", "1", "2"]
    cli.run()
    assert ran


def test_cli_no_annots():
    cli = Radicli()
    ran = False

    @cli.command("test")
    def test(a: str, b: int, c: float):
        assert a == "hello"
        assert b == 1
        assert c == 2.0
        nonlocal ran
        ran = True

    cli.run(["", "test", "hello", "1", "2"])
    assert ran


def test_cli_mix():
    cli = Radicli()
    ran = False

    @cli.command(
        "test",
        a=Arg(),
        b=Arg("--b", "-B"),
        c=Arg("--c", "-C"),
        d=Arg("--d", "-D"),
        e=Arg("--e", "-E"),
        f=Arg("--f", "-F"),
        g=Arg("--g", "-G"),
    )
    def test(a: str, b: int, c: float, d: bool, e: bool, f: bool, g: str = "yo"):
        assert a == "hello"
        assert b == 2
        assert c == 3.0
        assert d is True
        assert e is True
        assert f is False
        assert g == "yo"
        nonlocal ran
        ran = True

    cli.run(["", "test", "hello", "--b", "2", "-C", "3", "--d", "-E"])
    assert ran


def test_cli_lists():
    cli = Radicli()
    ran = False

    @cli.command("test", a=Arg("--a"), b=Arg("--b"), c=Arg("--c"))
    def test(a: str, b: List[str], c: Optional[List[int]] = None):
        assert a == "hello"
        assert b == ["one", "two"]
        assert c is None
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "hello", "--b", "one", "--b", "two"])
    assert ran


def test_cli_different_dest():
    cli = Radicli()
    ran = False

    @cli.command("test", first=Arg("--a"), second=Arg("--b"))
    def test(first: str, second: str):
        assert first == "one"
        assert second == "two"
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "one", "--b", "two"])
    assert ran


def test_cli_defaults():
    cli = Radicli()
    ran = False

    @cli.command("test", a=Arg(), b=Arg(), c=Arg("--c"), d=Arg("--d"))
    def test(a: str, b: str = "hey", *, c: List[str], d: Optional[List[int]] = None):
        assert a == "yo"
        assert b == "hey"
        assert c == ["one"]
        assert d is None
        nonlocal ran
        ran = True

    cli.run(["", "test", "yo", "--c", "one"])
    assert ran


def test_cli_required():
    cli = Radicli()

    @cli.command("test", a=Arg(), b=Arg("--b"), c=Arg("--c"), d=Arg("--d"))
    def test(a: str, b: str, c: int, d: int = 0):
        ...

    with pytest.raises(CliParserError) as err:
        cli.run(["", "test", "hello", "--c", "1"])
    assert str(err.value).endswith("required: --b")
    with pytest.raises(CliParserError) as err:
        cli.run(["", "test", "--b", "hello", "--c", "1"])
    assert str(err.value).endswith("required: a")
    with pytest.raises(CliParserError) as err:
        # Positional, so this is parsed in argparse before it hits custom logic
        cli.run(["", "test", "--c", "1"])
    assert str(err.value).endswith("required: a")
    with pytest.raises(CliParserError) as err:
        cli.run(["", "test", "hello", "--d", "1"])
    assert str(err.value).endswith("required: --b, --c")


def test_cli_literals():
    cli = Radicli()
    ran = False

    @cli.command("test", a=Arg("--a"), b=Arg("--b"))
    def test(a: Literal["pizza", "pasta"], b: Literal["cola", "fanta"]):
        assert a == "pizza"
        assert b == "fanta"
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "pizza", "--b", "fanta"])
    assert ran


def test_cli_literals_list():
    cli = Radicli()
    ran = False

    @cli.command("test", a=Arg("--a"))
    def test(a: List[Literal["pizza", "pasta", "burger"]]):
        assert a == ["pasta", "pizza"]
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "pasta", "--a", "pizza"])
    assert ran

    with pytest.raises(CliParserError):
        cli.run(["", "test", "--a", "burger", "--a", "fries"])


def test_cli_enums():
    cli = Radicli()
    ran = False

    class FoodEnum(Enum):
        pizza = "🍕"
        pasta = "🍝"
        burger = "🍔"

    class DrinkEnum(Enum):
        soda = "🥤"
        juice = "🧃"
        beer = "🍺"

    @cli.command("test", a=Arg("--a"), b=Arg("--b"))
    def test(a: FoodEnum, b: DrinkEnum):
        assert a == FoodEnum.burger
        assert b == DrinkEnum.beer
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "burger", "--b", "beer"])
    assert ran


@pytest.mark.parametrize(
    "args,count",
    [(["--verbose", "--verbose"], 2), (["-VVVVV"], 5)],
)
def test_cli_count(args, count):
    cli = Radicli()
    ran = False

    @cli.command("test", verbose=Arg("--verbose", "-V", count=True))
    def test(verbose: int):
        assert verbose == count
        nonlocal ran
        ran = True

    cli.run(["", "test", *args])
    assert ran


def test_cli_converter():
    cli = Radicli()
    converter = lambda x: x.upper()
    ran = False

    @cli.command("test", a=Arg("--a"), b=Arg("--b", converter=converter))
    def test(a: str, b: str):
        assert a == "hello"
        assert b == "WORLD"
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "hello", "--b", "world"])
    assert ran


def test_cli_invalid_converter():
    """Test that errors in converters aren't masked by argparse."""
    # Previously: argument --a: invalid converter value: 'hello'
    cli = Radicli()
    error_msg = "This is an error!"

    def converter(value):
        raise TypeError(error_msg)

    @cli.command("test", a=Arg("--a", converter=converter))
    def test(a: str):
        ...

    with pytest.raises(CliParserError, match=error_msg):
        cli.run(["", "test", "--a", "hello"])


def test_cli_global_converters():
    collected = []

    def convert_list(value: str):
        collected.append(value.upper())
        return collected

    @dataclass
    class CustomType:
        id: int
        name: str

    def convert_custom_type(value: str):
        c_id, name = value.split("|")
        return CustomType(id=int(c_id), name=name)

    converters = {
        str: lambda x: x.upper(),
        List[str]: convert_list,
        CustomType: convert_custom_type,
    }

    cli = Radicli(converters=converters)
    ran = False

    @cli.command("test", a=Arg("--a"), b=Arg("--b"), c=Arg("--c"))
    def test(a: str, b: List[str], c: CustomType):
        assert a == "HELLO"
        assert b == ["FOO", "BAR"]
        assert isinstance(c, CustomType)
        assert c.id == 123
        assert c.name == "Person"
        nonlocal ran
        ran = True

    args = ["", "test", "--a", "hello", "--b", "foo", "--b", "bar", "--c", "123|Person"]
    cli.run(args)
    assert ran


_KindT = TypeVar("_KindT", bound=Union[str, int, float, Path])


class CustomGeneric(Generic[_KindT]):
    ...


def test_cli_converters_generics():
    converters = {CustomGeneric: lambda value: f"generic: {value}"}
    cli = Radicli(converters=converters)
    ran = False

    @cli.command("test", a=Arg("--a"))
    def test(a: CustomGeneric[str]):
        assert a == "generic: x"
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "x"])
    assert ran


def test_cli_converters_generics_multiple():
    _KindT = TypeVar("_KindT")

    class CustomGeneric(Generic[_KindT]):
        ...

    converters = {
        CustomGeneric: lambda value: f"generic: {value}",
        CustomGeneric[str]: lambda value: f"generic str: {value}",
        CustomGeneric[int]: lambda value: f"generic int: {value}",
    }
    cli = Radicli(converters=converters)
    ran = False

    @cli.command("test", a=Arg("--a"), b=Arg("--b"), c=Arg("--c"), d=Arg("--d"))
    def test(
        a: CustomGeneric,
        b: CustomGeneric[Path],
        c: CustomGeneric[str],
        d: CustomGeneric[int],
    ):
        assert a == "generic: x"
        assert b == "generic: y"
        assert c == "generic str: z"
        assert d == "generic int: 3"
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "x", "--b", "y", "--c", "z", "--d", "3"])
    assert ran


def test_cli_with_extra():
    cli = Radicli()
    ran = False

    @cli.command_with_extra("test", a=Arg("--a"), b=Arg("--b"))
    def test(a: str, b: int, _extra: List[str]):
        assert a == "hello"
        assert b == 1
        assert _extra == ["--hello", "2", "--world"]
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "hello", "--b", "1", "--hello", "2", "--world"])
    assert ran


def test_cli_with_extra_custom_key():
    cli = Radicli(extra_key="additional")
    ran = False

    @cli.command_with_extra("test", a=Arg("--a"), b=Arg("--b"))
    def test(a: str, b: int, additional: List[str]):
        assert a == "hello"
        assert b == 1
        assert additional == ["--hello", "2", "--world"]
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "hello", "--b", "1", "--hello", "2", "--world"])
    assert ran


def test_cli_subcommands():
    cli = Radicli()
    ran_parent = False
    ran_child1 = False
    ran_child2 = False

    @cli.command("test", a=Arg("--a"), b=Arg("--b"))
    def test(a: int, b: str):
        # Base command to prevent triggering single-command use case
        ...

    @cli.command("parent", a=Arg("--a"), b=Arg("--b"))
    def parent(a: int, b: str):
        assert a == 1
        assert b == "hello"
        nonlocal ran_parent
        ran_parent = True

    @cli.subcommand("parent", "child1", a=Arg("--a"), b=Arg("--b"), c=Arg("--c"))
    def child1(a: str, b: int, c: bool):
        assert a == "hey"
        assert b == 2
        assert c
        nonlocal ran_child1
        ran_child1 = True

    @cli.subcommand("parent", "child2", x=Arg(), y=Arg("--y"))
    def child2(x: str, y: Literal["pizza", "pasta"]):
        assert x == "yo"
        assert y == "pasta"
        nonlocal ran_child2
        ran_child2 = True

    args_parent = ["--a", "1", "--b", "hello"]
    args_child1 = ["--a", "hey", "--b", "2", "--c"]
    args_child2 = ["yo", "--y", "pasta"]

    cli.run(["", "parent", *args_parent])
    assert ran_parent
    cli.run(["", "parent", "child1", *args_child1])
    assert ran_child1
    cli.run(["", "parent", "child2", *args_child2])
    assert ran_child2
    with pytest.raises(CommandNotFoundError):
        cli.run(["", "child1", *args_child1])
    with pytest.raises(CliParserError):
        cli.run(["", "parent", "child3"])
    with pytest.raises(CliParserError):
        cli.run(["", "parent", "child2", *args_child1])


def test_cli_subcommands_parent_extra():
    # Known limitation: extra arguments on parents with subcommands need to
    # be prefixed by - or --, otherwise they'll be falsely interpreted as a
    # subcommand.
    cli = Radicli()
    ran_parent = False
    ran_child = False

    @cli.command_with_extra("parent", a=Arg("--a"), b=Arg("--b"))
    def parent(a: int, b: str, _extra: List[str]):
        assert a == 1
        assert b == "hello"
        assert _extra == ["--xyz"]
        nonlocal ran_parent
        ran_parent = True

    @cli.subcommand("parent", "child", a=Arg("--a"), b=Arg("--b"))
    def child(a: str, b: int):
        assert a == "hey"
        assert b == 2
        nonlocal ran_child
        ran_child = True

    cli.run(["", "parent", "--a", "1", "--b", "hello", "--xyz"])
    assert ran_parent
    cli.run(["", "parent", "child", "--a", "hey", "--b", "2"])
    assert ran_child


def test_cli_subcommands_child_extra():
    cli = Radicli()
    ran_parent = False
    ran_child = False

    @cli.command("parent", a=Arg("--a"), b=Arg("--b"))
    def parent(a: int, b: str):
        assert a == 1
        assert b == "hello"
        nonlocal ran_parent
        ran_parent = True

    @cli.subcommand_with_extra("parent", "child", a=Arg("--a"), b=Arg("--b"))
    def child(a: str, b: int, _extra: List[str]):
        assert a == "hey"
        assert b == 2
        assert _extra == ["xyz"]
        nonlocal ran_child
        ran_child = True

    cli.run(["", "parent", "--a", "1", "--b", "hello"])
    assert ran_parent
    cli.run(["", "parent", "child", "--a", "hey", "--b", "2", "xyz"])
    assert ran_child


def test_cli_subcommands_no_parent():
    cli = Radicli()
    ran_child1 = False
    ran_child2 = False

    @cli.subcommand("parent", "child1", a=Arg("--a"), b=Arg("--b"), c=Arg("--c"))
    def child1(a: str, b: int, c: bool):
        assert a == "hey"
        assert b == 2
        assert c
        nonlocal ran_child1
        ran_child1 = True

    @cli.subcommand("parent", "child2", x=Arg(), y=Arg("--y"))
    def child2(x: str, y: Literal["pizza", "pasta"]):
        assert x == "yo"
        assert y == "pasta"
        nonlocal ran_child2
        ran_child2 = True

    cli.run(["", "parent", "child1", "--a", "hey", "--b", "2", "--c"])
    assert ran_child1
    cli.run(["", "parent", "child2", "yo", "--y", "pasta"])
    assert ran_child2


def test_cli_path_converters():
    cli = Radicli()
    dir_name = "my_dir"
    file_name = "my_file.txt"
    ran = False

    @cli.command("test", a=Arg("--a"), b=Arg("--b"), c=Arg("--c"))
    def test(a: ExistingPath, b: ExistingFilePath, c: ExistingDirPath):
        assert str(a) == str(file_path)
        assert str(b) == str(file_path)
        assert str(c) == str(dir_path)
        nonlocal ran
        ran = True

    with make_tempdir() as d:
        dir_path = d / dir_name
        dir_path.mkdir()
        file_path = d / file_name
        file_path.touch()
        bad_path = Path(d / "x.txt")

        args1 = ["--a", str(file_path), "--b", str(file_path), "--c", str(dir_path)]
        args2 = ["--a", str(bad_path), "--b", str(file_path), "--c", str(dir_path)]
        args3 = ["--a", str(file_path), "--b", str(dir_path), "--c", str(dir_path)]
        args4 = ["--a", str(file_path), "--b", str(file_path), "--c", str(file_path)]

        cli.run(["", "test", *args1])
        assert ran
        with pytest.raises(CliParserError):
            cli.run(["", "test", *args2])
        with pytest.raises(CliParserError):
            cli.run(["", "test", *args3])
        with pytest.raises(CliParserError):
            cli.run(["", "test", *args4])


def test_cli_path_or_dash():
    cli = Radicli()
    file_name = "my_file.txt"
    ran1 = False
    ran2 = False

    @cli.command("test1", a=Arg())
    def test1(a: ExistingFilePathOrDash):
        assert str(a) == str(file_path)
        nonlocal ran1
        ran1 = True

    @cli.command("test2", a=Arg())
    def test2(a: ExistingFilePathOrDash):
        assert a == "-"
        nonlocal ran2
        ran2 = True

    with make_tempdir() as d:
        file_path = d / file_name
        file_path.touch()
        bad_path = Path(d / "x.txt")

        cli.run(["", "test1", str(file_path)])
        assert ran1
        cli.run(["", "test2", "-"])
        assert ran2
        with pytest.raises(CliParserError):
            cli.run(["", "test1", str(bad_path)])
        with pytest.raises(CliParserError):
            cli.run(["", "test2", "_"])


def test_cli_stack_decorators():
    cli = Radicli()
    ran = 0

    @cli.command("one", a=Arg("--a"), b=Arg("--b"))
    @cli.command("two", a=Arg("--a"), b=Arg("--b"))
    @cli.command("three", a=Arg("--a"), b=Arg("--b"))
    def test(a: str, b: int):
        assert a == "hello"
        assert b == 1
        nonlocal ran
        ran += 1

    args = ["--a", "hello", "--b", "1"]
    cli.run(["", "one", *args])
    assert ran == 1
    cli.run(["", "two", *args])
    assert ran == 2
    cli.run(["", "three", *args])
    assert ran == 3


def test_cli_custom_help_arg():
    cli = Radicli()
    ran = False

    @cli.command("test", a=Arg("--a"), show_help=Arg("--help"))
    def test(a: str, show_help: bool):
        assert a == "hello"
        assert show_help
        nonlocal ran
        ran = True

    cli.run(["", "test", "--a", "hello", "--help"])
    assert ran


def test_cli_version(capsys):
    version = "1.2.3"
    cli = Radicli(version=version)
    ran = False

    @cli.command("test", a=Arg("--a"))
    def test(a: str):
        assert a == "hello"
        nonlocal ran
        ran = True

    with pytest.raises(SystemExit):
        cli.run(["", "--version"])
    captured = capsys.readouterr()
    assert captured.out.strip() == version
    cli.run(["", "test", "--a", "hello"])
    assert ran


def test_cli_version_multiple_commands(capsys):
    # Test --version also works on the top-level with no command specified
    version = "1.2.3"
    cli = Radicli(version=version)
    ran1 = False
    ran2 = False

    @cli.command("test1", a=Arg("--a"))
    def test1(a: str):
        nonlocal ran1
        ran1 = True

    @cli.command("test2", a=Arg("--a"))
    def test2(a: str):
        nonlocal ran2
        ran2 = True

    with pytest.raises(SystemExit):
        cli.run(["", "--version"])
    captured = capsys.readouterr()
    assert captured.out.strip() == version
    assert not ran1
    assert not ran2


def test_cli_single_command():
    """Test that the name can be left out for CLIs with only one command."""
    cli = Radicli()
    ran = False

    @cli.command("test", a=Arg("--a"))
    def test(a: str):
        assert a == "hello"
        nonlocal ran
        ran = True

    cli.run(["", "--a", "hello"])
    assert ran


def test_cli_single_command_subcommands():
    """Test that the name can be left out for CLIs with only one command."""
    cli = Radicli()
    ran_parent = False
    ran_child = False

    @cli.command("parent", a=Arg("--a"))
    def parent(a: str):
        assert a == "hello"
        nonlocal ran_parent
        ran_parent = True

    @cli.subcommand("parent", "child", a=Arg("--a"))
    def child(a: str):
        assert a == "hello"
        nonlocal ran_child
        ran_child = True

    cli.run(["", "--a", "hello"])
    assert ran_parent
    cli.run(["", "child", "--a", "hello"])
    assert ran_child


@pytest.mark.parametrize(
    "raise_error, handle_errors, handler_return, expect_handled, expect_exit",
    [
        (None, [KeyError], None, False, False),
        (KeyError, [KeyError], None, True, False),
        (KeyError, [KeyError], 1, True, True),
        (KeyError, [], 1, True, True),
        (KeyError, [ValueError], 1, True, True),
    ],
)
def test_cli_errors(
    raise_error: Optional[Type[Exception]],
    handle_errors: List[Type[Exception]],
    handler_return: Optional[int],
    expect_handled: bool,
    expect_exit: bool,
):
    handler_ran = False

    def error_handler(e: Exception) -> Optional[int]:
        nonlocal handler_ran
        handler_ran = True
        return handler_return

    cli = Radicli(errors={e: error_handler for e in handle_errors})

    @cli.command("test")
    def test():
        nonlocal ran
        ran = True
        if raise_error is not None:
            raise raise_error

    ran = False

    if raise_error is not None and raise_error not in handle_errors:
        with pytest.raises(raise_error):
            cli.run(["", "test"])
    elif expect_exit:
        with pytest.raises(SystemExit):
            cli.run(["", "test"])
    else:
        cli.run(["", "test"])
        assert ran
        assert handler_ran is expect_handled


def test_cli_static_roundtrip(capsys):
    cli = Radicli(prog="test")

    @cli.command("hello", a=Arg("--a", help="aaa"), b=Arg("--b", help="bbb"))
    def hello(a, b):
        """Hello"""
        ...

    @cli.command("world", c=Arg(help="ccc"))
    def world(c):
        """World"""
        ...

    with make_tempdir() as dir_path:
        path = dir_path / "static.json"
        cli.to_static(path)

        static = StaticRadicli.load(path)

    assert static.prog == cli.prog
    assert len(static.commands) == len(cli.commands)
    for parent, commands in static.subcommands:
        assert parent in cli.subcommands
        assert len(cli.subcommands[parent]) == len(commands)

    hello1 = static.commands["hello"]
    hello2 = cli.commands["hello"]
    assert hello1.name == hello2.name
    assert hello1.description == hello2.description
    for arg1, arg2 in zip(hello1.args, hello2.args):
        assert arg1.help == arg2.help
        assert arg1.arg.option == arg2.arg.option
        assert arg1.arg.short == arg2.arg.short
        assert arg1.arg.help == arg2.arg.help

    with pytest.raises(SystemExit):
        static.run(["", "--help"])
    captured1 = capsys.readouterr().out
    with pytest.raises(SystemExit):
        cli.run(["", "--help"])
    captured2 = capsys.readouterr().out
    assert captured1 == captured2

    with pytest.raises(SystemExit):
        static.run(["", "hello", "--help"])
    captured1 = capsys.readouterr().out
    with pytest.raises(SystemExit):
        cli.run(["", "hello", "--help"])
    captured2 = capsys.readouterr().out
    assert captured1 == captured2

    with pytest.raises(SystemExit):
        static.run(["", "world", "--help"])
    captured1 = capsys.readouterr().out
    with pytest.raises(SystemExit):
        cli.run(["", "world", "--help"])
    captured2 = capsys.readouterr().out
    assert captured1 == captured2

    with make_tempdir() as dir_path:
        path = dir_path / "static.json"
        cli.to_static(path)

        static = StaticRadicli.load(path, debug=True)

    with pytest.raises(SystemExit):
        static.run(["", "hello", "--help"])
    captured = capsys.readouterr().out
    assert static._debug_start in captured

    static.disable = True
    static.run(["", "hello", "--help"])
    captured = capsys.readouterr().out
    assert not captured


@pytest.mark.parametrize(
    "arg_type",
    [
        str,
        int,
        float,
        bool,
        List[str],
        Path,
        ExistingPath,
        ExistingFilePath,
        ExistingDirPath,
        ExistingFilePathOrDash,
        Literal["a", "b", "c"],
    ],
)
def test_static_deserialize_types(arg_type):
    """Test that supported and built-in types are correctly deserialized from static"""
    get_converter = lambda v: DEFAULT_CONVERTERS.get(v)
    arg = get_arg(
        "test", Arg("--test"), arg_type, orig_type=arg_type, get_converter=get_converter
    )
    arg_json = arg.to_static_json()
    new_arg = ArgparseArg.from_static_json(arg_json)
    assert new_arg.type == arg.type
    assert new_arg.orig_type == stringify_type(arg.orig_type)
    assert new_arg.has_converter == arg.has_converter
    assert new_arg.action == arg.action


@pytest.mark.parametrize(
    "arg_type",
    [
        List[str],
        ZipFile,
        Optional[ZipFile],
        Union[ZipFile, str, Path],
        CustomGeneric,
        CustomGeneric[int],
        CustomGeneric[str],
        Optional[CustomGeneric],
    ],
)
def test_static_deserialize_types_custom_deserialize(arg_type):
    """Test deserialization with custom type deserializer"""

    split_string = get_list_converter(str)

    def convert_zipfile(value: str) -> ZipFile:
        return ZipFile(value, "r")

    def convert_generic(value: str) -> str:
        return f"generic: {value}"

    converters = {
        ZipFile: convert_zipfile,  # new type with custom converter
        List[str]: split_string,
        CustomGeneric: convert_generic,
        CustomGeneric[str]: str,
    }
    get_converter = lambda v: converters.get(v)
    # With converters set
    arg = get_arg(
        "test", Arg("--test"), arg_type, orig_type=arg_type, get_converter=get_converter
    )
    arg_json = arg.to_static_json()
    new_arg = ArgparseArg.from_static_json(arg_json, converters=converters)
    assert new_arg.type == arg.type
    assert new_arg.orig_type == stringify_type(arg.orig_type)
    assert new_arg.has_converter == arg.has_converter
    assert new_arg.action == arg.action
    # With no converters set
    arg = get_arg(
        "test", Arg("--test"), arg_type, orig_type=arg_type, get_converter=get_converter
    )
    arg_json = arg.to_static_json()
    new_arg = ArgparseArg.from_static_json(arg_json)
    assert new_arg.type is str
    assert new_arg.orig_type == stringify_type(arg.orig_type)


@pytest.mark.parametrize(
    "arg_type,expected_type,expected_str",
    [
        (str, str, "str"),
        (List[str], List[str], "List[str]"),
        (Optional[List[str]], List[str], "List[str]"),
        (Union[str, int], str, "str"),
        (CustomGeneric, CustomGeneric, "CustomGeneric"),
        (CustomGeneric[str], CustomGeneric[str], "CustomGeneric[str]"),
        (ExistingFilePath, ExistingFilePath, "ExistingFilePath (Path)"),
        (ZipFile, ZipFile, "ZipFile"),
    ],
)
def test_cli_arg_display_type(arg_type, expected_type, expected_str):
    split_string = get_list_converter(str)

    def convert_zipfile(value: str) -> ZipFile:
        return ZipFile(value, "r")

    def convert_generic(value: str) -> str:
        return f"generic: {value}"

    converters = {
        **DEFAULT_CONVERTERS,
        ZipFile: convert_zipfile,  # new type with custom converter
        List[str]: split_string,
        CustomGeneric: convert_generic,
        CustomGeneric[str]: str,
    }

    def test(test: arg_type):
        ...

    cmd = Command.from_function(
        "test", {"test": Arg("--test")}, test, converters=converters
    )
    arg = cmd.args[0]
    assert arg.display_type == expected_type
    assert format_type(arg.display_type) == expected_str


def test_cli_no_defaults():
    cli = Radicli(fill_defaults=False)
    ran = False

    @cli.command(
        "test",
        a=Arg(),
        b=Arg(),
        c=Arg("--c", "-C"),
        d=Arg("--d", "-D"),
        e=Arg("--e", "-E"),
    )
    def test(a: str, b: str = "1", *, c: int = 1, d: bool = False, e: str = "yo"):
        assert a == "hello"
        assert b == "1"
        assert c == 3
        assert d is True
        assert e == "yo"
        nonlocal ran
        ran = True

    args = ["hello", "--c", "3", "--d"]
    parsed = cli.parse(args, cli.commands["test"])
    assert parsed == {"a": "hello", "c": 3, "d": True}
    cli.run(["", *args])
    assert ran
    # Make sure that set defaults are still preserved
    args = ["hello", "--c", "1", "--d"]
    parsed = cli.parse(args, cli.commands["test"])
    assert parsed == {"a": "hello", "c": 1, "d": True}



from typing import List, Union, Optional
from pathlib import Path
import json

from .cli import Radicli, Command
from .util import StaticData, ConvertersType, SimpleFrozenDict


class StaticRadicli(Radicli):
    data: StaticData
    disable: bool
    debug: bool

    def __init__(
        self,
        data: StaticData,
        disable: bool = False,
        debug: bool = False,
        converters: ConvertersType = SimpleFrozenDict(),
    ) -> None:
        super().__init__(
            prog=data["prog"],
            help=data["help"],
            version=data["version"],
            extra_key=data["extra_key"],
        )
        self.commands = {
            name: Command.from_static_json(cmd, converters)
            for name, cmd in data["commands"].items()
        }
        self.subcommands = {
            parent: {
                name: Command.from_static_json(sub, converters)
                for name, sub in subs.items()
            }
            for parent, subs in data["subcommands"].items()
        }
        self.data = data
        self.disable = disable
        self.debug = debug
        self._debug_start = "===== STATIC ====="
        self._debug_end = "=== END STATIC ==="

    def run(self, args: Optional[List[str]] = None) -> None:
        """
        Run the static CLI. Should usually happen before importing and running
        the live CLI so the static CLI can show help or raise errors and
        exit, without requiring importing the live CLI.
        """
        if self.disable:
            return
        if self.debug:
            print(self._debug_start)
        super().run(args)
        if self.debug:
            print(self._debug_end)

    @classmethod
    def load(
        cls,
        file_path: Union[str, Path],
        debug: bool = False,
        disable: bool = False,
        converters: ConvertersType = SimpleFrozenDict(),
    ) -> "StaticRadicli":
        """Load the static CLI from a file path created with Radicli.to_static."""
        path = Path(file_path)
        if not path.exists() or not path.is_file():
            raise ValueError(f"Not a valid file path: {path}")
        with path.open("r", encoding="utf8") as f:
            data = json.load(f)
        return cls(data, disable=disable, debug=debug, converters=converters)



import typing
from typing import Union, Generic, List, TypeVar
from pathlib import Path
import pathlib
from uuid import UUID
import pytest
import shutil
from radicli.util import stringify_type, get_list_converter

_KindT = TypeVar("_KindT", bound=Union[str, int, float, Path])


class CustomGeneric(Generic[_KindT]):
    ...


@pytest.mark.parametrize(
    "arg_type,expected",
    [
        (str, "str"),
        (bool, "bool"),
        (Path, "Path"),
        (List[int], "List[int]"),
        (CustomGeneric, "CustomGeneric"),
        (CustomGeneric[str], "CustomGeneric[str]"),
        (UUID, "UUID"),
        (shutil.rmtree, "rmtree"),
        (typing.List[pathlib.Path], "List[Path]"),
        (
            typing.Dict[Union[str, int], Union[str, pathlib.Path, int]],
            "Dict[Union[str, int], Union[str, Path, int]]",
        ),
        (Union[str, typing.Tuple[str, Path]], "Union[str, Tuple[str, Path]]"),
        ("foo.bar", "foo.bar"),
        (None, None),
    ],
)
def test_stringify_type(arg_type, expected):
    assert stringify_type(arg_type) == expected


@pytest.mark.parametrize(
    "item_type,value,expected",
    [
        # Separated string
        (str, "hello, world,test", ["hello", "world", "test"]),
        (int, " 1,2,3 ", [1, 2, 3]),
        (float, "0.123,5,  1.234", [0.123, 5.0, 1.234]),
        # Quoted list
        (str, "[hello, world]", ["hello", "world"]),
        (str, '["hello", "world"]', ["hello", "world"]),
        (str, "['hello', 'world']", ["hello", "world"]),
        (int, "[1,2,3]", [1, 2, 3]),
        (int, '["1","2","3"]', [1, 2, 3]),
        (int, "['1','2','3']", [1, 2, 3]),
        (float, "[0.23,2,3.45]", [0.23, 2.0, 3.45]),
        (float, '["0.23","2","3.45"]', [0.23, 2.0, 3.45]),
        (float, "['0.23','2','3.45']", [0.23, 2.0, 3.45]),
    ],
)
def test_get_list_converter(item_type, value, expected):
    converter = get_list_converter(item_type)
    assert converter(value) == expected



from typing import Any, Callable, Iterable, Type, Union, Optional, Dict, Tuple
from typing import List, Literal, NewType, get_args, get_origin, TypeVar
from typing import TypedDict, cast
from enum import Enum
from uuid import UUID
from dataclasses import dataclass
from pathlib import Path
import inspect
import argparse
import re

# We need this Iterable type, which is the type origin of types.Iterable
try:
    from collections.abc import Iterable as IterableType  # Python 3.9+
except ImportError:
    from collections import Iterable as IterableType  # type: ignore

DEFAULT_PLACEHOLDER = argparse.SUPPRESS
BASE_TYPES_MAP = {"str": str, "int": int, "float": float, "Path": Path}
BASE_TYPES = list(BASE_TYPES_MAP.values())
ConverterType = Callable[[str], Any]
ConvertersType = Dict[Union[Type, object], ConverterType]
ArgTypeType = Optional[Union[Type, ConverterType]]
_Exc = TypeVar("_Exc", bound=Exception, covariant=True)
ErrorHandlerType = Callable[[Exception], Optional[int]]
ErrorHandlersType = Dict[Type[_Exc], Callable[[_Exc], Optional[int]]]


class StaticArg(TypedDict):
    id: str
    option: Optional[str]
    short: Optional[str]
    orig_help: Optional[str]
    default: Optional[Union[bool, str]]
    help: Optional[str]
    action: Optional[str]
    choices: Optional[List[str]]
    has_converter: bool
    type: Optional[str]
    orig_type: Optional[str]


class StaticCommand(TypedDict):
    name: str
    args: List[StaticArg]
    description: Optional[str]
    allow_extra: bool
    parent: Optional[str]
    is_placeholder: bool


class StaticData(TypedDict):
    prog: Optional[str]
    help: str
    version: Optional[str]
    extra_key: str
    commands: Dict[str, StaticCommand]
    subcommands: Dict[str, Dict[str, StaticCommand]]


class SimpleFrozenDict(dict):
    """Simplified implementation of a frozen dict, mainly used as default
    function or method argument (for arguments that should default to empty
    dictionary).
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.error = "Can't write to frozen dict. This is likely an internal error."

    def __setitem__(self, key, value):
        raise NotImplementedError(self.error)

    def pop(self, key: Any, default=None):
        raise NotImplementedError(self.error)

    def update(self, other, **kwargs):
        raise NotImplementedError(self.error)


class CliParserError(SystemExit):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class UnsupportedTypeError(Exception):
    def __init__(self, arg: str, annot: Any) -> None:
        self.arg = arg
        self.annot = annot
        self.message = f"Unsupported type for '{self.arg}': {self.annot}"
        super().__init__(self.message)


class CommandNotFoundError(Exception):
    def __init__(self, name: str, options: List[str]) -> None:
        self.name = name
        self.options = options
        self.message = (
            f"Can't find command '{self.name}'. Available: {', '.join(self.options)}"
        )
        super().__init__(self.message)


class CommandExistsError(Exception):
    def __init__(self, name: str) -> None:
        self.name = name
        self.message = f"Command '{self.name}' already exists"
        super().__init__(self.message)


class InvalidArgumentError(Exception):
    def __init__(self, arg_id: str, message: str):
        self.id = arg_id
        self.msg = message
        self.message = f"Invalid argument '{self.id}': {self.msg}"
        super().__init__(self.message)


@dataclass
class Arg:
    """Field for defining the CLI argument in the decorator."""

    option: Optional[str] = None
    short: Optional[str] = None
    help: Optional[str] = None
    converter: Optional[ConverterType] = None
    count: bool = False


@dataclass
class ArgparseArg:
    """Internal argument dataclass defining values passed to argparse."""

    id: str
    arg: Arg
    type: ArgTypeType = None
    orig_type: Union[ArgTypeType, str] = None
    default: Any = DEFAULT_PLACEHOLDER
    # We modify the help to add types so we store it twice to store old and new
    help: Optional[str] = None
    action: Optional[Union[str, Type[argparse.Action]]] = None
    choices: Optional[Union[List[str], List[Enum]]] = None
    has_converter: bool = False

    @property
    def display_type(self) -> Union[ArgTypeType, str]:
        default_type = self.type if self.type is not None else self.orig_type
        return self.orig_type if self.has_converter else default_type

    def to_argparse(self) -> Tuple[List[str], Dict[str, Any]]:
        """Helper method to generate args and kwargs for Parser.add_argument."""
        args: List[str] = []
        if self.arg.option:
            args.append(self.arg.option)
        if self.arg.short:
            args.append(self.arg.short)
        kwargs: Dict[str, Any] = {
            "dest": self.id,
            "action": self.action,
            "help": self.help,
            "default": self.default,
        }
        # Support defaults for positional arguments
        if not self.arg.option and self.default is not DEFAULT_PLACEHOLDER:
            kwargs["nargs"] = "?"
        # Not all arguments are valid for all options
        if self.type is not None:
            kwargs["type"] = self.type
        if self.choices is not None:
            kwargs["choices"] = self.choices
        return args, kwargs

    def to_static_json(self) -> StaticArg:
        """Convert the argument to a JSON-serializable dict."""
        return {
            "id": self.id,
            "option": self.arg.option,
            "short": self.arg.short,
            "orig_help": self.arg.help,
            "default": str(self.default)
            if self.default not in (False, None)
            else self.default,
            "help": self.help,
            "action": str(self.action) if self.action else None,
            "choices": list(c.value if isinstance(c, Enum) else c for c in self.choices)
            if self.choices
            else None,
            "has_converter": self.has_converter,
            "type": stringify_type(self.type),
            "orig_type": stringify_type(self.orig_type),
        }

    @classmethod
    def from_static_json(
        cls,
        data: StaticArg,
        converters: ConvertersType = SimpleFrozenDict(),
    ) -> "ArgparseArg":
        """Initialize the static argument from a JSON-serializable dict."""
        return ArgparseArg(
            id=data["id"],
            arg=Arg(data["option"], data["short"], help=data["orig_help"]),
            type=deserialize_type(data, converters),
            orig_type=data["orig_type"],
            default=DEFAULT_PLACEHOLDER
            if data["default"] == DEFAULT_PLACEHOLDER
            else data["default"],
            help=data["help"],
            action=data["action"],
            choices=data["choices"],
            has_converter=data["has_converter"],
        )


def deserialize_type(
    data: StaticArg,
    converters: ConvertersType = SimpleFrozenDict(),
) -> ArgTypeType:
    # No type or special args with no type
    if data["type"] is None or data["action"] in ("store_true", "count"):
        return None
    # Handle custom types: we use the orig_type here, since this corresponds to
    # what was actually set as an argument type hint
    orig_type = data["orig_type"]
    if orig_type is None:
        return None
    converters_map = {stringify_type(k): v for k, v in converters.items()}
    if orig_type in converters_map:
        return converters_map[orig_type]
    # Hacky check for generics
    if "[" in orig_type:
        origin = orig_type.split("[", 1)[0]
        if origin in converters_map:
            return converters_map[orig_type.split("[", 1)[0]]
    # Check defaults last to honor custom converters for builtins
    types_map = {**BASE_TYPES_MAP}
    for value in DEFAULT_CONVERTERS.values():
        types_map[stringify_type(value)] = value  # type: ignore
    if data["type"] in types_map:
        return types_map[data["type"]]
    return str


def get_arg(
    param: str,
    orig_arg: Arg,
    param_type: Any,
    *,
    orig_type: Union[ArgTypeType, str] = None,
    default: Optional[Any] = DEFAULT_PLACEHOLDER,
    get_converter: Optional[Callable[[Type], Optional[ConverterType]]] = None,
    has_converter: bool = False,
    skip_resolve: bool = False,
) -> ArgparseArg:
    """Generate an argument to add to argparse and interpret types if possible."""
    arg = ArgparseArg(
        id=param,
        arg=orig_arg,
        type=param_type,
        help=orig_arg.help,
        default=default,
        orig_type=orig_type,
        has_converter=has_converter,
    )
    if orig_arg.count:
        arg.action = "count"
        arg.type = None
        if not arg.default:
            arg.default = 0
        return arg
    # Need to do this first so we can recursively resolve custom types like
    # Union[ExistingPath] etc.
    origin = get_origin(param_type)
    args = get_args(param_type)
    converter = get_converter(param_type) if get_converter else None
    if get_converter and not converter:
        # Check if we have a converter for the origin, e.g. for generics Foo[Bar]
        converter = get_converter(origin)  # type: ignore
    if converter:
        arg.type = converter
        arg.has_converter = True
        return arg
    if skip_resolve:
        return arg
    if origin is Union:
        if type(None) in args and default is DEFAULT_PLACEHOLDER:
            default = None
        arg_types = [a for a in args if a != type(None)]  # noqa: E721
        if arg_types:
            return get_arg(
                param,
                orig_arg,
                arg_types[0],
                orig_type=arg_types[0],
                default=default,
                get_converter=get_converter,
            )
    if param_type in BASE_TYPES:
        arg.type = param_type
        return arg
    if param_type is bool:
        if not orig_arg.option:
            raise InvalidArgumentError(
                arg.id,
                f"boolean arguments need to be flags, e.g. --{arg.id.replace('_', '-')}",
            )
        arg.type = None
        if default is True:
            raise InvalidArgumentError(arg.id, "boolean flags need to default to False")
        arg.default = False
        arg.action = "store_true"
        return arg
    if inspect.isclass(param_type) and issubclass(param_type, Enum):
        arg.choices = list(param_type.__members__.keys())
        arg.type = lambda value: getattr(param_type, value, value)
        return arg
    if not origin:
        raise UnsupportedTypeError(param, param_type)
    if origin is Literal and len(args):
        arg.choices = list(args)
        arg.type = type(args[0])
        return arg
    if origin in (list, IterableType):
        if len(args) and get_origin(args[0]) is Literal:
            literal_args = get_args(args[0])
            if literal_args:
                arg.type = type(literal_args[0])
                arg.choices = list(literal_args)
                arg.action = "append"
                return arg
        arg.type = find_base_type(args)
        arg.action = "append"
        return arg
    raise UnsupportedTypeError(param, param_type)


def find_base_type(
    args: Iterable[Any], default_type: Callable[[str], Any] = str
) -> Callable[[str], Any]:
    """Check a list of types for the next available basic type, e.g. str."""
    for base_type in BASE_TYPES:
        if base_type in args:
            return base_type
    return default_type


def stringify_type(arg_type: Any) -> Optional[str]:
    """Get a pretty-printed string for a type."""
    if isinstance(arg_type, str) or arg_type is None:
        return arg_type
    if hasattr(arg_type, "__name__"):
        type_str = arg_type.__name__
        args = get_args(arg_type)
        if args:
            # Built-in generic types are callables in Python 3.10+ so we want to
            # preserve args here and stringify them, too
            type_args = cast(List[str], [stringify_type(arg) for arg in args])
            type_str = f"{type_str}[{', '.join(type_args)}]"
        return type_str
    type_str = str(arg_type)
    return _stringify_type(str(arg_type))


subtype_matcher = re.compile(r"\[(.*)\]")


def _stringify_type(type_str: str) -> str:
    parts = []
    split_type = subtype_matcher.split(type_str)
    first = split_type.pop(0)
    split_first = first.rsplit(".", 1)
    parts.append(split_first[1] if len(split_first) == 2 else first)
    for substr in split_type:
        if substr:
            objs = [_stringify_type(sub.strip()) for sub in substr.split(",")]
            parts.extend(["[", ", ".join(objs), "]"])
    return "".join(parts)


def format_type(arg_type: Any) -> Optional[str]:
    """Get a pretty-printed string for a type."""
    type_str = stringify_type(arg_type)
    # Hacky check for cross-platform supertypes for NewType custom types
    if (
        (
            # Python < 3.10
            hasattr(arg_type, "__qualname__")
            and arg_type.__qualname__.startswith("NewType")
        )
        # Python 3.10+
        or (hasattr(arg_type, "__class__") and arg_type.__class__ is NewType)
    ) and (
        hasattr(arg_type, "__supertype__")
    ):  # type: ignore
        supertype = stringify_type(arg_type.__supertype__)  # type: ignore
        return f"{type_str} ({supertype})"
    return type_str


def join_strings(*strings: Optional[str], char: str = " ") -> str:
    return char.join(x for x in strings if x)


def format_table(data: List[Tuple[str, str]]) -> str:
    widths = [[len(str(col)) for col in item] for item in data]
    max_widths = [min(max(w), 50) for w in list(zip(*widths))]
    rows = []
    for item in data:
        cols = []
        for i, col in enumerate(item):
            cols.append(("{:%d}" % max_widths[i]).format(str(col or "")))
        rows.append((" " * 3).join(cols))
    return "\n" + "\n".join(rows) + "\n"


def format_arg_help(text: Optional[str], max_width: int = 70) -> str:
    if not text:
        return " "
    d = (text or "").strip()[:max_width]
    end = "." if "." in d or len(text or "") <= max_width else "..."
    return (d.rsplit(".", 1)[0] if "." in d else d) + end


def expand_error_subclasses(
    errors: Dict[Type[Exception], ErrorHandlerType]
) -> Dict[Type[Exception], ErrorHandlerType]:
    """Map subclasses of errors to their parent's handler."""
    output = {}
    for err, callback in errors.items():
        if hasattr(err, "__subclasses__"):
            for subclass in err.__subclasses__():
                output[subclass] = callback
        output[err] = callback
    return output


_InT = TypeVar("_InT", bound=Union[str, int, float])


def get_list_converter(
    type_func: Callable[[Any], _InT] = str, delimiter: str = ","
) -> Callable[[str], List[_InT]]:
    def converter(value: str) -> List[_InT]:
        if not value:
            return []
        if value.startswith("[") and value.endswith("]"):
            value = value[1:-1]
        result = []
        for p in value.split(delimiter):
            p = p.strip()
            if p.startswith("'") and p.endswith("'"):
                p = p[1:-1]
            if p.startswith('"') and p.endswith('"'):
                p = p[1:-1]
            p = type_func(p.strip())
            result.append(p)
        return result

    return converter


def convert_existing_path(path_str: str) -> Path:
    path = Path(path_str)
    if not path.exists():
        raise CliParserError(f"path does not exist: {path_str}")
    return path


def convert_existing_file_path(path_str: str) -> Path:
    path = convert_existing_path(path_str)
    if not path.is_file():
        raise CliParserError(f"path is not a file path: {path_str}")
    return path


def convert_existing_dir_path(path_str: str) -> Path:
    path = convert_existing_path(path_str)
    if not path.is_dir():
        raise CliParserError(f"path is not a directory path: {path_str}")
    return path


def convert_existing_path_or_dash(path_str: str) -> Union[Path, str]:
    if path_str == "-":
        return path_str
    return convert_existing_path(path_str)


def convert_existing_file_path_or_dash(path_str: str) -> Union[Path, str]:
    if path_str == "-":
        return path_str
    return convert_existing_file_path(path_str)


def convert_existing_dir_path_or_dash(path_str: str) -> Union[Path, str]:
    if path_str == "-":
        return path_str
    return convert_existing_dir_path(path_str)


def convert_path_or_dash(path_str: str) -> Union[Path, str]:
    if path_str == "-":
        return path_str
    return Path(path_str)


def convert_uuid(value: str) -> UUID:
    return UUID(value)


def convert_str_or_uuid(value: str) -> Union[str, UUID]:
    try:
        return UUID(value)
    except ValueError:
        return value


# Custom path types for custom converters
ExistingPath = NewType("ExistingPath", Path)
ExistingFilePath = NewType("ExistingFilePath", Path)
ExistingDirPath = NewType("ExistingDirPath", Path)

ExistingPathOrDash = Union[ExistingPath, Literal["-"]]
ExistingFilePathOrDash = Union[ExistingFilePath, Literal["-"]]
ExistingDirPathOrDash = Union[ExistingDirPath, Literal["-"]]
PathOrDash = Union[Path, Literal["-"]]
StrOrUUID = Union[str, UUID]


DEFAULT_CONVERTERS: ConvertersType = {
    ExistingPath: convert_existing_path,
    ExistingFilePath: convert_existing_file_path,
    ExistingDirPath: convert_existing_dir_path,
    ExistingPathOrDash: convert_existing_path_or_dash,
    ExistingFilePathOrDash: convert_existing_file_path_or_dash,
    ExistingDirPathOrDash: convert_existing_dir_path_or_dash,
    PathOrDash: convert_path_or_dash,
    UUID: convert_uuid,
    StrOrUUID: convert_str_or_uuid,
}



from typing import List, Iterable, Optional, Union, Literal, Dict, Any
from enum import Enum
import pytest
import argparse
from radicli import Radicli, Arg, Command
from radicli.util import get_arg, UnsupportedTypeError, CliParserError


class FoodEnum(Enum):
    pizza = "🍕"
    pasta = "🍝"
    burger = "🍔"


GOOD_TEST_CASES = [
    (
        ["--a", "1", "--b", "2", "--c", "2"],
        [
            get_arg("a", Arg("--a"), str),
            get_arg("b", Arg("--b"), int),
            get_arg("c", Arg("--c"), float),
        ],
        {"a": "1", "b": 2, "c": 2.0},
    ),
    # Positional arguments
    (
        ["1", "2", "--c", "2"],
        [
            get_arg("a", Arg(), str),
            get_arg("b", Arg(), int),
            get_arg("c", Arg("--c"), float),
        ],
        {"a": "1", "b": 2, "c": 2.0},
    ),
    # Booleans
    (["--a"], [get_arg("a", Arg("--a"), bool)], {"a": True}),
    (
        ["--a", "1", "--b", "--c", "3"],
        [
            get_arg("a", Arg("--a"), str),
            get_arg("b", Arg("--b"), bool),
            get_arg("c", Arg("--c"), int),
            get_arg("d", Arg("--d"), bool),
        ],
        {"a": "1", "b": True, "c": 3, "d": False},
    ),
    # List types and iterables
    (
        ["--a", "1", "--a", "2", "--a", "3"],
        [get_arg("a", Arg("--a"), List[str])],
        {"a": ["1", "2", "3"]},
    ),
    (
        ["--a", "1", "--a", "2", "--a", "3"],
        [get_arg("a", Arg("--a"), List[int])],
        {"a": [1, 2, 3]},
    ),
    (
        ["--a", "1", "--a", "2", "--a", "3"],
        [get_arg("a", Arg("--a"), List[float])],
        {"a": [1.0, 2.0, 3.0]},
    ),
    (
        ["--a", "1", "--a", "2", "--a", "3"],
        [get_arg("a", Arg("--a"), Iterable[str])],
        {"a": ["1", "2", "3"]},
    ),
    (
        ["--a", "1", "--a", "2", "--a", "3"],
        [get_arg("a", Arg("--a"), Optional[List[str]])],
        {"a": ["1", "2", "3"]},
    ),
    # Optional arguments
    (
        ["--a", "1", "--c", "3"],
        [
            get_arg("a", Arg("--a"), Optional[str]),
            get_arg("b", Arg("--b"), Optional[int]),
            get_arg("c", Arg("--c"), Union[str, int]),
        ],
        {"a": "1", "b": None, "c": "3"},
    ),
    # Shorthand format
    (
        ["-A", "1", "--b", "2", "-C", "2"],
        [
            get_arg("a", Arg("--a", "-A"), str),
            get_arg("b", Arg("--b", "-B"), int),
            get_arg("c", Arg("--c", "-C"), float),
        ],
        {"a": "1", "b": 2, "c": 2.0},
    ),
    # Custom converter
    (
        ["--a", "hello world"],
        [get_arg("a", Arg("--a"), lambda x: x.upper(), skip_resolve=True)],
        {"a": "HELLO WORLD"},
    ),
    # Literals
    (
        ["--a", "pizza", "--b", "fanta"],
        [
            get_arg("a", Arg("--a"), Literal["pizza", "pasta", "burger"]),
            get_arg("b", Arg("--b"), Literal["cola", "fanta", "sprite"]),
        ],
        {"a": "pizza", "b": "fanta"},
    ),
    # Enums
    (
        ["--a", "pizza"],
        [get_arg("a", Arg("--a"), FoodEnum)],
        {"a": FoodEnum.pizza},
    ),
    # Counting
    (
        ["--verbose", "--verbose"],
        [get_arg("verbose", Arg("--verbose", "-v", count=True), int)],
        {"verbose": 2},
    ),
    (
        ["-vvv"],
        [get_arg("verbose", Arg("--verbose", "-v", count=True), int)],
        {"verbose": 3},
    ),
]

EXTRA_KEY = "__extra__"
GOOD_WITH_EXTRA_TEST_CASES = [
    (
        ["--a", "1", "--b", "2", "--hello", "3", "--world"],
        [get_arg("a", Arg("--a"), str), get_arg("b", Arg("--b"), int)],
        {"a": "1", "b": 2, EXTRA_KEY: ["--hello", "3", "--world"]},
    ),
]

GOOD_PARTIAL_TEST_CASES = [
    (
        ["--a", "1"],
        [get_arg("a", Arg("--a"), str), get_arg("b", Arg("--b"), int)],
        {"a": "1"},
    ),
]

BAD_TEST_CASES = [
    # Unsupported types
    (
        ["--a", "{'hello': 'world'}"],
        [("a", Arg("--a"), Dict[str, Any])],
        UnsupportedTypeError,
    ),
    # Bad values
    (
        ["--a", "hello"],
        [("a", Arg("--a"), int)],
        CliParserError,
    ),
    # Unrecognized, missing or duplicate arguments
    (["--a", "1", "--b", "2"], [("a", Arg("--a"), str)], CliParserError),
    (["--a", "1"], [("a", Arg(), str)], CliParserError),
    (["--b", "1"], [("a", Arg(), str), ("b", Arg("--b"), str)], CliParserError),
    (
        ["--a", "1"],
        [("a", Arg("--a"), str), ("a", Arg("--a"), str)],
        argparse.ArgumentError,
    ),
    (
        ["--a", "1", "--b", "2"],
        [("a", Arg("--a", "-A"), str), ("b", Arg("--b", "-A"), str)],
        argparse.ArgumentError,
    ),
    # Literals
    (
        ["--a", "fries"],
        [("a", Arg("--a"), Literal["pizza", "pasta", "burger"])],
        CliParserError,
    ),
    # Enums
    (["--a", "fries"], [("a", Arg("--a"), FoodEnum)], CliParserError),
]


@pytest.mark.parametrize(
    "args,arg_info,expected",
    GOOD_TEST_CASES,
)
def test_parser_good(args, arg_info, expected):
    cli = Radicli()
    cmd = Command(name="test", func=lambda: None, args=arg_info)
    assert cli.parse(args, cmd) == expected


@pytest.mark.parametrize(
    "args,arg_info,expected",
    GOOD_WITH_EXTRA_TEST_CASES,
)
def test_parser_good_with_extra(args, arg_info, expected):
    cli = Radicli(extra_key=EXTRA_KEY)
    cmd = Command(name="test", func=lambda: None, args=arg_info, allow_extra=True)
    assert cli.parse(args, cmd) == expected


@pytest.mark.parametrize(
    "args,arg_info,expected",
    GOOD_PARTIAL_TEST_CASES,
)
def test_parser_good_partial(args, arg_info, expected):
    cli = Radicli()
    cmd = Command(name="test", func=lambda: None, args=arg_info)
    assert cli.parse(args, cmd, allow_partial=True) == expected


@pytest.mark.parametrize(
    "args,get_arg_args,expected_error",
    BAD_TEST_CASES,
)
def test_parser_bad(args, get_arg_args, expected_error):
    cli = Radicli()
    with pytest.raises(expected_error):
        arg_info = [get_arg(*args) for args in get_arg_args]
        cmd = Command(name="test", func=lambda: None, args=arg_info)
        cli.parse(args, cmd)
